/**
* License: 2-clause BSD
*
* Copyright (c) 2016, Martin Belanger <nitram_67@hotmail.com>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, this
*    list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*/

#include <string.h>   /* memset() */
#include "config.h"   /* HAVE_DIGEST_LENGTH_ARG */
#include "rc-hmac.h"

void hmac_md5_with_nettle(uint8_t *data, size_t  data_len,
                          uint8_t *key,  size_t  key_len,
                          uint8_t  digest[MD5_DIGEST_SIZE])
{
    struct hmac_md5_ctx md5;
    memset(digest, 0, MD5_DIGEST_SIZE);
    hmac_md5_set_key(&md5, key_len, key);
    hmac_md5_update(&md5, data_len, data);
#ifdef HAVE_DIGEST_LENGTH_ARG
    hmac_md5_digest(&md5, MD5_DIGEST_SIZE, digest);
#else
    hmac_md5_digest(&md5, digest);
#endif
}

